from typing import List, Iterable, Type
from time import sleep
import acme.challenges
from acme.challenges import ChallengeResponse, Challenge
from certbot import interfaces
from certbot.achallenges import AnnotatedChallenge
from certbot.plugins import common
from certbot.display import util as display_util
from .rox_api import RoxApi


class Authenticator(common.Plugin, interfaces.Authenticator):
    description = "Obtain certificates using a DNS TXT record (if you are using rockenstein AG " \
                  "for DNS). "

    def __init__(self, *args, **kwargs):
        super(Authenticator, self).__init__(*args, **kwargs)
        self._roxApi = RoxApi(self.conf('token'), self.conf('url'), not self.conf('ignore-ssl'))

    def prepare(self) -> None:
        return None

    @classmethod
    def add_parser_arguments(cls, add) -> None:
        super(Authenticator, cls).add_parser_arguments(
            add
        )
        add("token", help="Token for rockenstein API")
        add("url", help="URL for rockenstein API", default="https://api.rox.net")
        add("ignore-ssl", action="store_true", default=False)
        add('propagation-seconds',
            default=120,
            type=int,
            help='The number of seconds to wait for DNS to propagate before asking the ACME server '
                 'to verify the DNS record.')

    def more_info(self) -> str:
        return self.description

    def perform(self, achalls: List[AnnotatedChallenge]) -> List[ChallengeResponse]:
        responses = []
        for achall in achalls:
            domain = achall.domain
            validation_domain_name = achall.validation_domain_name(domain)
            validation = achall.validation(achall.account_key)

            # domains can be subdomains, eg. "abc.example.com" so we have to find out
            # the base domain name
            basedomain = self._roxApi.get_base_domain(domain)

            # make dns entry
            self._roxApi.add_txt_record(basedomain + '.',
                                        validation_domain_name + '.',
                                        validation)

            responses.append(achall.response(achall.account_key))

        # wait for DNS
        display_util.notify("Waiting %d seconds for DNS changes to propagate" %
                            self.conf('propagation-seconds'))
        sleep(self.conf('propagation-seconds'))

        return responses

    def cleanup(self, achalls: List[AnnotatedChallenge]) -> None:
        responses = []
        for achall in achalls:
            domain = achall.domain
            validation_domain_name = achall.validation_domain_name(domain)
            validation = achall.validation(achall.account_key)

            # domains can be subdomains, eg. "abc.example.com" so we have to find out
            # the base domain name
            basedomain = self._roxApi.get_base_domain(domain)

            # delete dns entry
            self._roxApi.del_txt_record(basedomain + '.',
                                        validation_domain_name + '.',
                                        validation)

            responses.append(achall.response(achall.account_key))

    def get_chall_pref(self, domain: str) -> Iterable[Type[Challenge]]:
        return [
            acme.challenges.DNS01
        ]



import requests
import json
from certbot import errors
from certbot.plugins import dns_common


class RoxApi:
    _token = ""
    _url = ""
    _verify_ssl = True

    def __init__(self, token, url, verify_ssl=True):
        self._token = token
        self._url = url
        self._verify_ssl = verify_ssl

    def _header(self):
        return {
            'X-AUTH-TOKEN': self._token
        }

    def _getTxtEntries(self, zonename, record_name):
        tmp = []
        response = requests.get(self._url + "/api/dns/zone/" + zonename,
                                headers=self._header(),
                                verify=self._verify_ssl)
        zone = json.loads(response.content)
        rrsets = zone["rrsets"]
        for rrset in rrsets:
            if rrset["name"] == record_name and rrset["type"] == "TXT":
                records = rrset["records"]
                for record in records:
                    tmp.append(record["content"].strip('"'))
        return tmp

    def add_txt_record(self, zonename, record_name, record_content, record_ttl=60):

        records = []
        entries = self._getTxtEntries(zonename, record_name)
        entries.append(record_content)
        for entry in entries:
            record = {
                "content": "\"" + entry + "\""
            }
            if not (record in records):
                records.append(record)

        record = {
            "name": record_name,
            "type": "TXT",
            "ttl": record_ttl,
            "changetype": "REPLACE",
            "records": records
        }

        response = requests.patch(self._url + "/api/dns/zone/" + zonename,
                                  data=json.dumps(record),
                                  verify=self._verify_ssl,
                                  headers=self._header())

        if 200 <= response.status_code < 300:
            return

        raise errors.PluginError('Error while creating dns records: {}.'.format(response.text))

    def del_txt_record(self, zonename, record_name, record_content, record_ttl=60):

        records = []
        entries = self._getTxtEntries(zonename, record_name)
        entries.append(record_content)
        for entry in entries:
            record = {
                "content": "\"" + entry + "\""
            }
            if not (record in records) and entry != record_content:
                records.append(record)

        if len(records) > 0:
            changetype = "REPLACE"
        else:
            changetype = "DELETE"

        record = {
            "name": record_name,
            "type": "TXT",
            "ttl": record_ttl,
            "changetype": changetype,
            "records": records
        }

        response = requests.patch(self._url + "/api/dns/zone/" + zonename,
                                  data=json.dumps(record),
                                  verify=self._verify_ssl,
                                  headers=self._header())

        if 200 <= response.status_code < 300:
            return

        raise errors.PluginError('Error removing dns records: {}.'.format(response.text))

    def get_base_domain(self, subdomain):
        subdomain_parts = dns_common.base_domain_name_guesses(subdomain)

        for subdomain_part in subdomain_parts:
            response = requests.get(self._url + "/api/dns/zone/" + subdomain_part + '.',
                                    headers=self._header(),
                                    verify=self._verify_ssl)

            if response.status_code == 200:
                return subdomain_part



from setuptools import setup, find_packages

setup(
    name='certbot-dns-rockenstein',
    version='0.0.5',
    description="rockenstein DNS Authenticator plugin for Certbot",
    url='https://www.rockenstein.de',
    author="Frank Mueller",
    author_email='fm@rockenstein.de',
    install_requires=[
        'certbot',
        'requests'
    ],
    entry_points={
        'certbot.plugins': [
            'certbot-dns-rockenstein = certbot_dns_rockenstein.certbot_dns_rockenstein:Authenticator'
        ],
    },
    packages=find_packages(),
)
